"""Custom :ref:`Celery subcommands <celery:extending-subcommands>`.

The subcommand that is implemented in :samp:`gwcelery.tools.{submodule}` can be
invoked from the command line as :samp:`gwcelery {submodule}`.
"""
